"""
Built-in constraints

All classes need a transform class. Note, unlike sklearn, transform can copy
or overwrite input depending on copy attribute.
"""

from abc import (ABC as _ABC, abstractmethod as _abstractmethod)

import numpy as _np

__all__ = ['Constraint', 'ConstraintNonneg', 'ConstraintCumsumNonneg',
           'ConstraintZeroEndPoints', 'ConstraintZeroCumSumEndPoints',
           'ConstraintNorm', 'ConstraintCutBelow', 'ConstraintCutAbove',
           'ConstraintCompressBelow', 'ConstraintCutAbove',
           'ConstraintCompressAbove', 'ConstraintReplaceZeros',
           'ConstraintPlanarize']


class Constraint(_ABC):
    """ Abstract class for constraints

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, copy=True):
        self.copy = copy

    @_abstractmethod
    def transform(self, A):
        """ Transform A input based on constraint """


class ConstraintNonneg(Constraint):
    """
    Non-negativity constraint. All negative entries made 0.

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """
    def __init__(self, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)

    def transform(self, A):
        """ Apply nonnegative constraint"""
        if self.copy:
            return A*(A > 0)
        else:
            A *= (A > 0)
            return A

class ConstraintCumsumNonneg(Constraint):
    """
    Cumulative-Summation non-negativity constraint. All negative
     entries made 0.

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """
    def __init__(self, axis=-1, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)
        self.axis = axis

    def transform(self, A):
        """ Apply cumsum nonnegative constraint"""
        
        if self.copy:
            return A*(_np.cumsum(A, self.axis) > 0)
        else:
            A *= (_np.cumsum(A, self.axis) > 0)
            return A
        

class ConstraintZeroEndPoints(Constraint):
    """
    Enforce the endpoints (or the mean over a range) is zero

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)

    axis : int
        Axis to operate on

    span : int
        Number of pixels along the ends to average.
    """
    def __init__(self, axis=-1, span=1, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)
        if [0, 1, -1].count(axis) != 1:
            raise TypeError('Axis must be 0, 1, or -1')

        self.axis = axis
        self.span = span

    def transform(self, A):
        """ Apply cumsum nonnegative constraint"""

        pix_vec = _np.arange(A.shape[self.axis])
        if (self.axis == 0):
            if self.span == 1:
                slope = (A[-1, :] - A[0, :]) / (pix_vec[-1] - pix_vec[0])
                intercept = A[0, :]
            else:
                slope = ((A[-self.span:, :].mean(axis=0) -
                        A[:self.span, :].mean(axis=0)) /
                        (pix_vec[-self.span:].mean() -
                        pix_vec[:self.span].mean()))
                intercept = (A[:self.span, :] -
                            _np.dot(pix_vec[:self.span, None],
                                    slope[None, :])).mean(axis=0)
            if self.copy:
                return A - _np.dot(pix_vec[:, None],
                                slope[None, :]) - intercept[None, :]
            else:
                A -= (_np.dot(pix_vec[:, None],
                            slope[None, :]) + intercept[None, :])
                return A
        else:
            if self.span == 1:
                slope = (A[:, -1] - A[:, 0]) / (pix_vec[-1] - pix_vec[0])
                intercept = A[:, 0]
            else:
                slope = ((A[:, -self.span:].mean(axis=1) -
                        A[:, :self.span].mean(axis=1)) /
                        (pix_vec[-self.span:].mean() -
                        pix_vec[:self.span].mean()))
                intercept = (A[:, :self.span] -
                            _np.dot(slope[:, None],
                                    pix_vec[None, :self.span])).mean(axis=1)

            if self.copy:
                return A - _np.dot(slope[:, None],
                                pix_vec[None, :]) - intercept[:, None]
            else:
                A -= (_np.dot(slope[:, None],
                            pix_vec[None, :]) + intercept[:, None])
                return A

class ConstraintZeroCumSumEndPoints(Constraint):
    """
    Enforce the endpoints of the cumsum (or the mean over a range) is
    near-zero. Note: this is an approximation.

    Parameters
    ----------
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    nodes : list of int
        In addition to end-points, other points to ensure are approximately 0
    axis : int
        Axis to operate on

    span : int
        Number of pixels along the ends to average.
    """
    def __init__(self, nodes=None, axis=-1, copy=False):
        """ A must be non-negative"""
        super().__init__(copy)
        
        self.nodes = nodes
        if [0, 1, -1].count(axis) != 1:
            raise TypeError('Axis must be 0, 1, or -1')

        self.axis = axis

    def transform(self, A):
        """ Apply cumsum nonnegative constraint"""
        
        meaner = A.mean(self.axis)

        if self.nodes:
            self.nodes = set(self.nodes)
            self.nodes.update({0, A.shape[self.axis]})
            self.nodes.discard(-1)
            self.nodes = list(self.nodes)

            if (self.axis == 0):
                if self.copy:
                    temp = 1.0*A
                    for num in range(len(self.nodes) - 1):
                        n0 = self.nodes[num]
                        n1 = self.nodes[num+1]
                        temp[n0:n1, :] -= temp[n0:n1, :].mean(self.axis)[None, :]
                    return temp
                else:
                    for num in range(len(self.nodes) - 1):
                        n0 = self.nodes[num]
                        n1 = self.nodes[num+1]
                        A[n0:n1, :] -= A[n0:n1, :].mean(self.axis)[None, :]
                    return A
            else:
                if self.copy:
                    temp = 1.0*A
                    for num in range(len(self.nodes) - 1):
                        n0 = self.nodes[num]
                        n1 = self.nodes[num+1]
                        temp[:, n0:n1] -= temp[:, n0:n1].mean(self.axis)[:, None]
                    return temp
                else:
                    for num in range(len(self.nodes) - 1):
                        n0 = self.nodes[num]
                        n1 = self.nodes[num+1]
                        A[:, n0:n1] -= A[:, n0:n1].mean(self.axis)[:, None]
                    return A
        else:
            if (self.axis == 0):
                if self.copy:
                    return A - meaner[None, :]
                else:
                    A -= meaner[None, :]
                    return A
            else:
                if self.copy:
                    return A - meaner[:, None]
                else:
                    A -= meaner[:, None]
                    return A
      

class ConstraintNorm(Constraint):
    """
    Normalization constraint.

    Parameters
    ----------
    axis : int
        Which axis of input matrix A to apply normalization acorss.
    fix : list
        Keep fix-axes as-is and normalize the remaining axes based on the
        residual of the fixed axes.
    set_zeros_to_feature : int
        Set all samples which sum-to-zero across axis to 1 for a particular
         feature (See Notes)
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)

    Notes
    -----

    -   For set_zeros_to_feature, assuming the data represents concentration
         with a matrix [n_samples, n_features] and the axis is across the
         features, for every sample that sums to 0 across axis, would be
         replaced with a vector [n_features] of zeros except at
         set_zeros_to_feature, which would equal 1. I.e., this pixel is
         now pure substance of index value set_zeros_to_feature.


    """
    def __init__(self, axis=-1, fix=None, copy=False):
        """Normalize along axis"""
        super().__init__(copy)
        if fix is None:
            self.fix = fix
        elif isinstance(fix, int):
            self.fix = [fix]
        elif isinstance(fix, (list, tuple)):
            self.fix = fix
        elif isinstance(fix, _np.ndarray):
            if _np.issubdtype(fix.dtype, _np.integer):
                self.fix = fix.tolist()
            else:
                raise TypeError('ndarrays must be of dtype int')
        else:
            raise TypeError('Parameter fix must be of type None, int, list,',
                            'tuple, ndarray')

        if not ((axis == 0) | (axis == 1) | (axis == -1)):
            raise ValueError('Axis must be 0,1, or -1')
        self.axis = axis

    def transform(self, A):
        """ Apply normalization constraint """
        
        if self.copy:
            if self.axis == 0:
                if not self.fix:  # No fixed axes
                    return A / A.sum(axis=self.axis)[None, :]
                else:  # Fixed axes
                    fix_locs = self.fix
                    not_fix_locs = [v for v in _np.arange(A.shape[0]).tolist()
                                    if self.fix.count(v) == 0]
                    scaler = _np.ones(A.shape)
                    div = A[not_fix_locs, :].sum(axis=0)[None, :]
                    div[div == 0] = 1
                    scaler[not_fix_locs, :] = ((1 - A[fix_locs, :].sum(axis=0)[None, :]) / div)

                    return A * scaler
            else:  # Axis = 1 / -1
                if not self.fix:  # No fixed axes
                    return A / A.sum(axis=self.axis)[:, None]
                else:  # Fixed axis
                    fix_locs = self.fix
                    not_fix_locs = [v for v in _np.arange(A.shape[-1]).tolist()
                                    if self.fix.count(v) == 0]
                    scaler = _np.ones(A.shape)
                    div = A[:, not_fix_locs].sum(axis=-1)[:, None]
                    div[div == 0] = 1
                    scaler[:, not_fix_locs] = ((1 - A[:, fix_locs].sum(axis=-1)[:, None]) / div)

                    return A * scaler
        else:  # Overwrite original data
            if A.dtype != _np.float:
                raise TypeError('A.dtype must be float for',
                                'in-place math (copy=False)')

            if self.axis == 0:
                if not self.fix:  # No fixed axes
                    A /= A.sum(axis=self.axis)[None, :]
                    return A
                else:  # Fixed axes
                    fix_locs = self.fix
                    not_fix_locs = [v for v in _np.arange(A.shape[0]).tolist()
                                    if self.fix.count(v) == 0]
                    scaler = _np.ones(A.shape)
                    div = A[not_fix_locs, :].sum(axis=0)[None, :]
                    div[div == 0] = 1
                    scaler[not_fix_locs, :] = ((1 - A[fix_locs, :].sum(axis=0)[None, :]) / div)
                    A *= scaler
                    return A
            else:  # Axis = 1 / -1
                if not self.fix:  # No fixed axes
                    A /= A.sum(axis=self.axis)[:, None]
                    return A
                else:  # Fixed axis
                    fix_locs = self.fix
                    not_fix_locs = [v for v in _np.arange(A.shape[-1]).tolist()
                                    if self.fix.count(v) == 0]
                    scaler = _np.ones(A.shape)
                    div = A[:, not_fix_locs].sum(axis=-1)[:, None]
                    div[div == 0] = 1
                    scaler[:, not_fix_locs] = ((1 - A[:, fix_locs].sum(axis=-1)[:, None]) / div)
                    A *= scaler
                    return A


class ConstraintReplaceZeros(Constraint):
    """
    Samples that sum-to-zero across axis are replaced with a vector of 0's except
    for a 1 at feature if a single value. In a concentration context, e.g., samples with
    no concentration are replaced with 100% concentration of a set feature. If multiple
    features given, equal amounts of each feature (summing to 1) are used.

    Parameters
    ----------
    axis : int
        Which axis of input matrix A to apply normalization acorss.
    feature : int, list, tuple
        Set all samples which sum-to-zero across axis to fval for a particular
         feature (or fractional) for multiple features.
    fval : float
        Value of summation across axis of replacement vector.
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)

    """

    def __init__(self, axis=-1, feature=None, fval=1, copy=False):
        """Replace sum-to-zero samples with new feature vector along axis"""
        super().__init__(copy)
        self.fval = fval
        if feature is None:
            self.feature = feature
        elif isinstance(feature, int):
            self.feature = [feature]
        elif isinstance(feature, (list, tuple)):
            self.feature = feature
        elif isinstance(feature, _np.ndarray):
            if _np.issubdtype(feature.dtype, _np.integer):
                self.feature = feature.tolist()
            else:
                raise TypeError('ndarrays must be of dtype int')
        else:
            raise TypeError('Parameter feature must be of type None, int, list, tuple, ndarray')

        if not ((axis == 0) | (axis == 1) | (axis == -1)):
            raise ValueError('Axis must be 0,1, or -1')
        self.axis = axis

    def transform(self, A):
        """ Apply constraint """

        if self.feature:
            replacement = _np.zeros(A.shape[self.axis])
            replacement[self.feature] = self.fval
            replacement /= replacement.sum()
            replacement *= self.fval

            if self.copy:
                A_out = 1*A
                if self.axis == 0:
                    A_out[:, A_out.sum(axis=0) == 0] = replacement[:, None]
                else:  # Axis 1 / -1
                    A_out[A_out.sum(axis=-1) == 0] = replacement
                return A_out
            else:
                if self.axis == 0:
                    A[:, A.sum(axis=0) == 0] = replacement[:, None]
                else:  # Axis 1 / -1
                    A[A.sum(axis=-1) == 0] = replacement
                return A
        else:
            return A
        

class _CutExclude(Constraint):
    """
    Parent class for methods that cut and can exclude

    Parameters
    ----------

    value : float
        Cutoff value
    axis_sumnz : int
        If not None, cut below value only applied where sum across specified
        axis does not go to 0, i.e. all values cut.
    exclude : int, list , tuple, ndarray
        Exclude targets
    exclude_axis : int
        Along which axis to enumerate targets
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """
    def __init__(self, value=0, axis_sumnz=None, exclude=None,
                 exclude_axis=-1, copy=False):
        """ """
        super().__init__(copy)
        self.value = value
        self.axis = axis_sumnz
        self.exclude = exclude
        self.exclude_axis = exclude_axis

        self._excl_mat = None

    def _make_excl_mat(self, A_shape):
        X, Y = _np.meshgrid(_np.arange(A_shape[1]), _np.arange(A_shape[0]))
        if self.exclude is None:
            self._excl_mat = _np.zeros(X.shape, dtype=_np.bool)
        else:
            if self.exclude_axis == 0:
                self._excl_mat = _np.in1d(Y.ravel(), self.exclude).reshape(Y.shape)
            else:
                self._excl_mat = _np.in1d(X.ravel(), self.exclude).reshape(X.shape)


class ConstraintCutBelow(_CutExclude):
    """
    Cut values below (and not-equal to) a certain threshold.

    Parameters
    ----------

    value : float
        Cutoff value
    axis_sumnz : int
        If not None, cut below value only applied where sum across specified
        axis does not go to 0, i.e. all values cut.
    exclude : int, list , tuple, ndarray
        Exclude targets
    exclude_axis : int
        Along which axis to enumerate targets
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """
    def __init__(self, value=0, axis_sumnz=None, exclude=None, exclude_axis=-1, copy=False):
        """ Initialize """
        super().__init__(value=value, axis_sumnz=axis_sumnz, exclude=exclude,
                         exclude_axis=exclude_axis, copy=copy)

    def transform(self, A):
        """ Apply cut-below value constraint"""
        if self._excl_mat is None:
            self._make_excl_mat(A.shape)

        if self.axis is None:
            if self.copy:
                return A*((A >= self.value) | self._excl_mat)
            else:
                A *= ((A >= self.value) | self._excl_mat)
                return A
        else:
            if self.copy:
                return A*(_np.alltrue(A < self.value, axis=self.axis, keepdims=True) +
                        (A >= self.value) + self._excl_mat)
            else:
                A *= (_np.alltrue(A < self.value, axis=self.axis, keepdims=True) +
                    (A >= self.value) + self._excl_mat)
                return A

class ConstraintCompressBelow(Constraint):
    """
    Compress values below (and not-equal to) a certain threshold (set to value)

    Parameters
    ----------

    value : float
        Cutoff value
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, value=0, copy=False):
        """  """
        super().__init__(copy)
        self.value = value

    def transform(self, A):
        """ Apply compress-below value constraint"""

        if self.copy:
            return A*(A >= self.value) + self.value*(A < self.value)
        else:
            temp = self.value*(A < self.value)
            A *= (A >= self.value)
            A += temp
            return A


class ConstraintCutAbove(_CutExclude):
    """
    Cut values above (and not-equal to) a certain threshold

    Parameters
    ----------

    value : float
        Cutoff value
    axis_sumnz : int
        If not None, cut above value only applied where sum across specified
        axis does not go to 0, i.e. all values cut.
    exclude : int, list , tuple, ndarray
        Exclude targets
    exclude_axis : int
        Along which axis to enumerate targets
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """
    def __init__(self, value=0, axis_sumnz=None, exclude=None, exclude_axis=-1, copy=False):
        """ """
        super().__init__(value=value, axis_sumnz=axis_sumnz, exclude=exclude,
                         exclude_axis=exclude_axis, copy=copy)

    def transform(self, A):
        """ Apply cut-above value constraint"""

        if self._excl_mat is None:
            self._make_excl_mat(A.shape)
        if self.axis is None:
            if self.copy:
                return A*((A <= self.value) | self._excl_mat)
            else:
                A *= ((A <= self.value) | self._excl_mat)
                return A
        else:
            if self.copy:
                return A*(_np.alltrue(A > self.value, axis=self.axis, keepdims=True) +
                        (A <= self.value) + self._excl_mat)
            else:
                A *= (_np.alltrue(A > self.value, axis=self.axis, keepdims=True) +
                    (A <= self.value) + self._excl_mat)
                return A


class ConstraintCompressAbove(Constraint):
    """
    Compress values above (and not-equal to) a certain threshold (set to value)

    Parameters
    ----------

    value : float
        Cutoff value
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)
    """

    def __init__(self, value=0, copy=False):
        """  """
        super().__init__(copy)
        self.value = value

    def transform(self, A):
        """ Apply compress-above value constraint"""
        
        if self.copy:
            return A*(A <= self.value) + self.value*(A > self.value)
        else:
            temp = self.value*(A > self.value)
            A *= (A <= self.value)
            A += temp
            return A
        
class ConstraintPlanarize(Constraint):
    """
    Set a particular target to a plane

    Parameters
    ----------

    target : int, list, tuple
        Target numbers to set to a fitted plane
    shape : tuple, list
        Shape of array (M,N) which is (Y,X)
    use_vals_above : float
        Only calculate based on values above (not including)
    use_vals_below : float
        Only calculate based on values below (not including)
    lims_to_plane : bool
        The returned plane will be limited to the range of the optionally supplied
        use_vals_below, use_vals above.
    scaler : float
        A large value that is much bigger than any values in the input array.
        Needed to ensure SVD properly creates plane. If None, auto-calculates.
    recalc_scaler : bool
        Auto-calculate for every new input (does not use previously provided or
        calculated value)
    copy : bool
        Make copy of input data, A; otherwise, overwrite (if mutable)

    Notes
    -----

    -   This uses an SVD to calculate the vector normal to the plane that fits the input data. It
        assumes that the 3rd singular vector is the normal; thus, the x and y vectors for the data need
        be larger than the variance of the input data. Scaler enables this by scaling the auto-generated
        x and y vectors to be much larger than the max-min of the input data

    """

    def __init__(self, target, shape, use_vals_above=None, use_vals_below=None,
                 lims_to_plane=True, scaler=None, recalc_scaler=False, copy=False):
        super().__init__(copy)
        if isinstance(target, int):
            self.target = [target]
        elif isinstance(target, (list, tuple, _np.ndarray)):
            self.target = target
        else:
            raise TypeError('target must be an int, list, 2D ndarray, or tuple')

        self.shape = shape
        self.copy = copy
        self.scaler = scaler
        self.recalc = recalc_scaler
        self.use_above = use_vals_above
        self.use_below = use_vals_below
        self.lims_to_plane = lims_to_plane

        self._x = None
        self._y = None
        self._X = None
        self._Y = None

        if scaler is not None:
            self._setup_xy(scaler)

    def _setup_xy(self, scaler):

        self.scaler = scaler
        self._x = scaler*_np.arange(self.shape[1], dtype=_np.float)
        self._y = scaler*_np.arange(self.shape[0], dtype=_np.float)

        self._X, self._Y = _np.meshgrid(self._x, self._y)
        self._X = self._X.ravel()
        self._Y = self._Y.ravel()

    def transform(self, A):
        """ Set targets, t, to fit planes """
        
        if (self.scaler is None) | (self.recalc):
            self._setup_xy(1e3 * _np.abs(A.max() - A.min()))

        if self.copy:
            A2 = 1*A

        for t in self.target:
            X2 = 1*self._X
            Y2 = 1*self._Y
            Z2 = 1*A[:, t]

            Stack = _np.vstack((X2, Y2, Z2))

            if self.use_above is not None:
                X2 = X2[Z2 > self.use_above]
                Y2 = Y2[Z2 > self.use_above]
                Z2 = Z2[Z2 > self.use_above]
                Stack = _np.vstack((X2, Y2, Z2))

            if self.use_below is not None:
                X2 = X2[Z2 < self.use_below]
                Y2 = Y2[Z2 < self.use_below]
                Z2 = Z2[Z2 < self.use_below]
                Stack = _np.vstack((X2, Y2, Z2))

            Stack = Stack - Stack.mean(axis=-1)[:, None]

            U, s, Vh = _np.linalg.svd(Stack, full_matrices=False)
            norm_to_plane = 1*U[:, -1]

            plane = (((-norm_to_plane[0] * (self._X - X2.mean())) -
                    (norm_to_plane[1] * (self._Y - Y2.mean()))) /
                    norm_to_plane[2]) + Z2.mean()

            if self.lims_to_plane:
                if self.use_above is not None:
                    plane[plane < self.use_above] = self.use_above
                if self.use_below is not None:
                    plane[plane > self.use_below] = self.use_below
            if not self.copy:
                A[:, t] = plane
            else:
                A2[:, t] = plane
        if self.copy:
            return A2
        else:
            return A
        

if __name__ == '__main__':  # pragma: no cover
    A = _np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(_np.float)
    A_transform = _np.array([[1, 2, 3, 4], [4, 0, 0, 0], [0, 0, 0, 0]]).astype(_np.float)

    constr = ConstraintCutAbove(copy=True, value=4)
    out = constr.transform(A)

    from numpy.testing import assert_allclose as _assert_allclose
    _assert_allclose(out, A_transform)



""" Functions to condition / preprocess data """
import numpy as _np

__all__ = ['standardize']

def standardize(X, mean_ctr=True, with_std=True, axis=-1, copy=True):
    """
    Standardization of data
    
    Parameters
    ----------

    X : ndarray
        Data array

    mean_ctr : bool
        Mean-center data

    with_std : bool
        Normalize by the standard deviation of the data
    
    axis : int
        Axis from which to calculate mean and standard deviation

    copy : bool
        Copy data (X) if True, overwite if False

    """

    if copy:
        Xsc = 1*X
    else:
        Xsc = X

    if mean_ctr:
        Xsc -= X.mean(axis=axis, keepdims=True)
    if with_std:
        Xsc /= X.std(axis=axis, keepdims=True)
    return Xsc


"""
Testing pymcr.regressors

"""

import numpy as np

from numpy.testing import (assert_equal, assert_array_equal, 
                           assert_allclose)

from pymcr.metrics import mse

def test_mse():
    A = np.ones((3,3))
    B = np.eye(3)
    mse(None, None, A, B)

# import pytest  


"""
Testing pymcr.constraints

"""

import numpy as np

from numpy.testing import assert_allclose

from pymcr.constraints import (ConstraintNonneg, ConstraintNorm,
                               ConstraintCumsumNonneg, ConstraintZeroEndPoints,
                               ConstraintZeroCumSumEndPoints, ConstraintCompressAbove,
                               ConstraintCompressBelow, ConstraintCutAbove, ConstraintCutBelow,
                               ConstraintReplaceZeros, ConstraintPlanarize)

import pytest

def test_nonneg():
    A = np.array([[1, 2, 3], [-1, -2, -3], [1, 2, 3]])
    A_nn = np.array([[1, 2, 3], [0, 0, 0], [1, 2, 3]])

    constr_nn = ConstraintNonneg(copy=True)
    out = constr_nn.transform(A)
    assert_allclose(A_nn, out)

    constr_nn = ConstraintNonneg(copy=False)
    out = constr_nn.transform(A)
    assert_allclose(A_nn, A)

def test_cumsumnonneg():
    """ Cum-Sum Nonnegativity Constraint """
    A = np.array([[2, -2, 3, -2], [-1, -2, -3, 7], [1, -2, -3, 7]])
    A_nn_ax1 = np.array([[2, 0, 3, -2], [0, 0, 0, 7], [1, 0, 0, 7]])
    A_nn_ax0 = np.array([[2, 0, 3, 0], [-1, 0, 0, 7], [1, 0, 0, 7]])

    # Axis -1
    constr_nn = ConstraintCumsumNonneg(copy=True, axis=-1)
    out = constr_nn.transform(A)
    assert_allclose(A_nn_ax1, out)

    # Axis 0
    constr_nn = ConstraintCumsumNonneg(copy=False, axis=0)
    out = constr_nn.transform(A)
    assert_allclose(A_nn_ax0, A)

def test_zeroendpoints():
    """ 0-Endpoints Constraint """
    A = np.array([[1, 2, 3, 4], [3, 6, 9, 12], [4, 8, 12, 16]]).astype(np.float)
    A_ax1 = np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]).astype(np.float)
    A_ax0 = np.array([[0, 0, 0, 0], [0.5, 1, 1.5, 2], [0, 0, 0, 0]]).astype(np.float)

    # Axis 0
    constr_ax0 = ConstraintZeroEndPoints(copy=True, axis=0)
    out = constr_ax0.transform(A)
    assert_allclose(A_ax0, out)

    # Axis -1
    constr_ax1 = ConstraintZeroEndPoints(copy=True, axis=-1)
    out = constr_ax1.transform(A)
    assert_allclose(A_ax1, out)

    with pytest.raises(TypeError):
        constr_ax1 = ConstraintZeroEndPoints(copy=True, axis=3)

    # Axis 0 -- NOT copies
    constr_ax0 = ConstraintZeroEndPoints(copy=False, axis=0)
    out = constr_ax0.transform(A)
    assert_allclose(A_ax0, A)

def test_zeroendpoints_span():
    """ 0-Endpoints Constraint """
    A = np.array([[1, 2, 3, 4], [3, 6, 9, 12], [4, 8, 12, 16]]).astype(np.float)

    # Axis 1
    constr_ax1 = ConstraintZeroEndPoints(copy=True, axis=1, span=2)
    out = constr_ax1.transform(A)
    assert_allclose(out[:, [0,1]].mean(axis=1), 0)
    assert_allclose(out[:, [1,2]].mean(axis=1), 0)

    # Axis 0
    constr_ax0 = ConstraintZeroEndPoints(copy=True, axis=0, span=2)
    out = constr_ax0.transform(A)
    assert_allclose(out[[0,1], :].mean(axis=0), 0)
    assert_allclose(out[[1,2], :].mean(axis=0), 0)

    # effective an assert_not_equal
    assert_allclose([q != 0 for q in out[:,0]], True)
    assert_allclose([q != 0 for q in out[:,-1]], True)

    # Axis 1 -- no copy
    constr_ax1 = ConstraintZeroEndPoints(copy=False, axis=1, span=2)
    out = constr_ax1.transform(A)
    assert_allclose(A[:, [0,1]].mean(axis=1), 0)
    assert_allclose(A[:, [1,2]].mean(axis=1), 0)

def test_zerocumsumendpoints():
    """ Cum-Sum 0-Endpoints Constraint """
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_diff1 = np.array([[1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]]).astype(np.float)
    A_diff0 = np.array([[3, 3, 3, 3], [3, 3, 3, 3], [3, 3, 3, 3]]).astype(np.float)

    # A_ax1 = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    # A_ax0 = np.array([[0, 0, 0], [0.5, 1, 1.5], [0, 0, 0]])

    # Axis 0
    constr_ax0 = ConstraintZeroCumSumEndPoints(copy=True, axis=0)
    out = constr_ax0.transform(A_diff0)
    assert_allclose(out, 0)
    assert_allclose(np.cumsum(out, axis=0), 0)

    # Axis -1
    constr_ax1 = ConstraintZeroCumSumEndPoints(copy=True, axis=-1)
    out = constr_ax1.transform(A_diff1)
    assert_allclose(out, 0)
    assert_allclose(np.cumsum(out, axis=1), 0)

    # Axis = -1 -- NOT copy
    constr_ax1 = ConstraintZeroCumSumEndPoints(copy=False, axis=-1)
    out = constr_ax1.transform(A_diff1)
    assert_allclose(A_diff1, 0)
    assert_allclose(np.cumsum(A_diff1, axis=1), 0)

def test_zerocumsumendpoints_nodes():
    """ Cum-Sum 0-Endpoints Constraint with defined nodes"""
    
    A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(np.float)
    A_transform_ax0 = np.array([[0, 0, -1, 5], [0, 0, 0, -4], [0, 0, 1, -1]])
    A_transform_ax1 = np.array([[-3, -2, -3, 8], [-1, 0, 0, 1], [-2, -1, 0, 3]])

    # COPY Axis 0, node=0 (same as endpoints)
    constr_ax0 = ConstraintZeroCumSumEndPoints(copy=True, nodes=[0], axis=0)
    out = constr_ax0.transform(A)
    assert_allclose(out, A_transform_ax0)
    # assert_allclose(np.cumsum(out, axis=0), 0)

    # COPY Axis -1, node=0 (same as endpoints)
    constr_ax1 = ConstraintZeroCumSumEndPoints(copy=True, nodes=[0], axis=-1)
    out = constr_ax1.transform(A)
    assert_allclose(out, A_transform_ax1)
    # assert_allclose(np.cumsum(out, axis=1), 0)

    # OVERWRITE, Axis = 0, node=0 (same as endpoints)
    A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(np.float)
    constr_ax0 = ConstraintZeroCumSumEndPoints(copy=False, nodes=[0], axis=0)
    out = constr_ax0.transform(A)
    assert_allclose(A, A_transform_ax0)

     # OVERWRITE, Axis = 1, node=0 (same as endpoints)
    A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(np.float)
    constr_ax1 = ConstraintZeroCumSumEndPoints(copy=False, nodes=[0], axis=-1)
    out = constr_ax1.transform(A)
    assert_allclose(A, A_transform_ax1)

    # COPY, Axis = 0, Nodes = all
    A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(np.float)
    constr_ax0 = ConstraintZeroCumSumEndPoints(copy=True, nodes=[0,1,2], axis=0)
    out = constr_ax0.transform(A)
    assert_allclose(out, 0)

    # COPY, Axis = 1, Nodes = all
    A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(np.float)
    constr_ax1 = ConstraintZeroCumSumEndPoints(copy=True, nodes=[0,1,2,3], axis=1)
    out = constr_ax1.transform(A)
    assert_allclose(out, 0)

    # OVERWRITE, Axis = 0, Nodes = all
    A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(np.float)
    constr_ax0 = ConstraintZeroCumSumEndPoints(copy=False, nodes=[0,1,2], axis=0)
    out = constr_ax0.transform(A)
    assert_allclose(A, 0)

    # OVERWRITE, Axis = 1, Nodes = all
    A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(np.float)
    constr_ax1 = ConstraintZeroCumSumEndPoints(copy=False, nodes=[0,1,2,3], axis=1)
    out = constr_ax1.transform(A)
    assert_allclose(A, 0)

def test_norm():
    """ Test normalization """
    # A must be dtype.float for in-place math (copy=False)
    constr_norm = ConstraintNorm(axis=0, copy=False)
    A = np.array([[1, 2, 3], [-1, -2, -3], [1, 2, 3]]) # dtype: int32
    with pytest.raises(TypeError):
        out = constr_norm.transform(A)

    # Axis must be 0,1, or -1
    with pytest.raises(ValueError):
        constr_norm = ConstraintNorm(axis=2, copy=False)

    A = np.array([[1, 2, 3], [-1, -2, -3], [1, 2, 3]], dtype=np.float)
    A_norm0 = A / A.sum(axis=0)[None,:]
    A_norm1 = A / A.sum(axis=1)[:,None]

    constr_norm = ConstraintNorm(axis=0, copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_norm0, out)

    constr_norm = ConstraintNorm(axis=1, copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_norm1, out)

    constr_norm = ConstraintNorm(axis=-1, copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_norm1, out)

    constr_norm = ConstraintNorm(axis=0, copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_norm0, A)

    A = np.array([[1, 2, 3], [-1, -2, -3], [1, 2, 3]], dtype=np.float)
    constr_norm = ConstraintNorm(axis=1, copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_norm1, A)

def test_norm_fixed_axes():
    # AXIS = 1
    A = np.array([[0.0, 0.2, 1.0, 0.0], [0.25, 0.25, 0.0, 0.0], [0.3, 0.9, 0.6, 0.0]], dtype=np.float)
    A_fix2_ax1 = np.array([[0.0, 0.0, 1.0, 0.0], [0.5, 0.5, 0.0, 0.0], [0.1, 0.3, 0.6, 0.0]], dtype=np.float)

    # Fixed axes must be integers
    with pytest.raises(TypeError):
        constr_norm = ConstraintNorm(axis=1, fix=2.2, copy=True)

    # Dtype must be integer related
    with pytest.raises(TypeError):
        constr_norm = ConstraintNorm(axis=1, fix=np.array([2.2]), copy=True)

    # COPY: True
    # Fix of type int
    constr_norm = ConstraintNorm(axis=1, fix=2, copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # Fix of type list
    constr_norm = ConstraintNorm(axis=1, fix=[2, 3], copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # Fix of type tuple
    constr_norm = ConstraintNorm(axis=1, fix=(2), copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # Fix of type ndarray
    constr_norm = ConstraintNorm(axis=1, fix=np.array([2]), copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # COPY: False
    A_fix2_ax1 = np.array([[0.0, 0.0, 1.0, 0.0], [0.5, 0.5, 0.0, 0.0], [0.1, 0.3, 0.6, 0.0]], dtype=np.float)

    # Fix of type int
    A = np.array([[0.0, 0.2, 1.0, 0.0], [0.25, 0.25, 0.0, 0.0], [0.3, 0.9, 0.6, 0.0]], dtype=np.float)
    constr_norm = ConstraintNorm(axis=1, fix=2, copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

    # Fix of type list
    A = np.array([[0.0, 0.2, 1.0, 0.0], [0.25, 0.25, 0.0, 0.0], [0.3, 0.9, 0.6, 0.0]], dtype=np.float)
    constr_norm = ConstraintNorm(axis=1, fix=[2,3], copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

    # Fix of type tuple
    A = np.array([[0.0, 0.2, 1.0, 0.0], [0.25, 0.25, 0.0, 0.0], [0.3, 0.9, 0.6, 0.0]], dtype=np.float)
    constr_norm = ConstraintNorm(axis=1, fix=(2), copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

    # Fix of type ndarray
    A = np.array([[0.0, 0.2, 1.0, 0.0], [0.25, 0.25, 0.0, 0.0], [0.3, 0.9, 0.6, 0.0]], dtype=np.float)
    constr_norm = ConstraintNorm(axis=1, fix=np.array([2]), copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

    # AXIS = 0
    # Lazy, so just transposed
    A = np.array([[0.0, 0.2, 1.0, 0.0], [0.25, 0.25, 0.0, 0.0], [0.3, 0.9, 0.6, 0.0]], dtype=np.float).T
    A_fix2_ax1 = np.array([[0.0, 0.0, 1.0, 0.0], [0.5, 0.5, 0.0, 0.0], [0.1, 0.3, 0.6, 0.0]], dtype=np.float).T
    # COPY: True
    # Fix of type int
    constr_norm = ConstraintNorm(axis=0, fix=2, copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # Fix of type list
    constr_norm = ConstraintNorm(axis=0, fix=[2,3], copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # Fix of type tuple
    constr_norm = ConstraintNorm(axis=0, fix=(2), copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # Fix of type ndarray
    constr_norm = ConstraintNorm(axis=0, fix=np.array([2]), copy=True)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, out)

    # COPY: False
    A_fix2_ax1 = np.array([[0.0, 0.0, 1.0], [0.5, 0.5, 0.0], [0.1, 0.3, 0.6]], dtype=np.float).T

    # Fix of type int
    A = np.array([[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=np.float).T
    constr_norm = ConstraintNorm(axis=0, fix=2, copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

    # Fix of type list
    A = np.array([[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=np.float).T
    constr_norm = ConstraintNorm(axis=0, fix=[2], copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

    # Fix of type tuple
    A = np.array([[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=np.float).T
    constr_norm = ConstraintNorm(axis=0, fix=(2), copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

    # Fix of type ndarray
    A = np.array([[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=np.float).T
    constr_norm = ConstraintNorm(axis=0, fix=np.array([2]), copy=False)
    out = constr_norm.transform(A)
    assert_allclose(A_fix2_ax1, A)

def test_cut_below():
    """ Test cutting below (and not equal to) a value """
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_transform = np.array([[0, 0, 0, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)

    constr = ConstraintCutBelow(copy=True, value=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform)

    # No Copy
    constr = ConstraintCutBelow(copy=False, value=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform)

def test_cut_below_exclude():
    """ Test cutting below (and not equal to) a value """
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_transform_excl_0_ax1 = np.array([[1, 0, 0, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_transform_excl_0_ax0 = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)

    # COPY
    constr = ConstraintCutBelow(copy=True, value=4, exclude=0, exclude_axis=1)
    out = constr.transform(A)
    assert_allclose(out, A_transform_excl_0_ax1)

    constr = ConstraintCutBelow(copy=True, value=4, exclude=0, exclude_axis=0)
    out = constr.transform(A)
    assert_allclose(out, A_transform_excl_0_ax0)

    # OVERWRITE
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    constr = ConstraintCutBelow(copy=False, value=4, exclude=0, exclude_axis=1)
    out = constr.transform(A)
    assert_allclose(A, A_transform_excl_0_ax1)

    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    constr = ConstraintCutBelow(copy=False, value=4, exclude=0, exclude_axis=0)
    out = constr.transform(A)
    assert_allclose(A, A_transform_excl_0_ax0)

def test_cut_below_nonzerosum():
    """
    Test cutting below (and not equal to) a value with the constrain that
        no columns (axis=-1) will all become 0
    """
    A = np.array([[0.3, 0.7], [0.4, 0.6], [0.5, 0.5]])
    cutoff = 0.7
    A_correct = np.array([[0.0, 0.7], [0.4, 0.6], [0.5, 0.5]])

    constr = ConstraintCutBelow(copy=True, value=cutoff, axis_sumnz=-1)
    out = constr.transform(A)
    assert_allclose(out, A_correct)

    constr = ConstraintCutBelow(copy=False, value=cutoff, axis_sumnz=-1)
    constr.transform(A)
    assert_allclose(A, A_correct)

def test_cut_below_nonzerosum_exclude():
    """
    Test cutting below (and not equal to) a value with the constrain that
        no columns (axis=-1) will all become 0
    """
    A = np.array([[0.3, 0.7], [0.3, 0.7], [0.7, 0.3]])
    cutoff = 0.7
    A_correct = np.array([[0.3, 0.7], [0.3, 0.7], [0.7, 0.0]])

    # COPY
    constr = ConstraintCutBelow(copy=True, value=cutoff, axis_sumnz=-1,
                                exclude=0, exclude_axis=-1)
    out = constr.transform(A)
    assert_allclose(out, A_correct)

    # OVERWRITE
    constr = ConstraintCutBelow(copy=False, value=cutoff, axis_sumnz=-1,
                                exclude=0, exclude_axis=-1)
    _ = constr.transform(A)
    assert_allclose(A, A_correct)

    # constr = ConstraintCutBelow(copy=False, value=cutoff, axis_sumnz=-1)
    # constr.transform(A)
    # assert_allclose(A, A_correct)

def test_cut_above_nonzerosum():
    """
    Test cutting above (and not equal to) a value with the constrain that
        no columns (axis=-1) will all become 0
    """
    A = np.array([[0.3, 0.7], [0.4, 0.6], [0.5, 0.5]])
    cutoff = 0.4
    A_correct = np.array([[0.3, 0.0], [0.4, 0.0], [0.5, 0.5]])

    constr = ConstraintCutAbove(copy=True, value=cutoff, axis_sumnz=-1)
    out = constr.transform(A)
    assert_allclose(out, A_correct)

    # NO Copy
    A = np.array([[0.3, 0.7], [0.4, 0.6], [0.5, 0.5]])
    cutoff = 0.4
    A_correct = np.array([[0.3, 0.0], [0.4, 0.0], [0.5, 0.5]])

    constr = ConstraintCutAbove(copy=False, value=cutoff, axis_sumnz=-1)
    constr.transform(A)
    assert_allclose(A, A_correct)

def test_compress_below():
    """ Test compressing below (and not equal to) a value """
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_transform = np.array([[4, 4, 4, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)

    constr = ConstraintCompressBelow(copy=True, value=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform)

    # No Copy
    constr = ConstraintCompressBelow(copy=False, value=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform)

def test_cut_above():
    """ Test cutting above (and not equal to) a value """
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_transform = np.array([[1, 2, 3, 4], [4, 0, 0, 0], [0, 0, 0, 0]]).astype(np.float)

    constr = ConstraintCutAbove(copy=True, value=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform)

    # No Copy
    constr = ConstraintCutAbove(copy=False, value=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform)

def test_cut_above_exclude():
    """ Test cutting above (and not equal to) a value """
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_transform_excl_0_ax1 = np.array([[1, 2, 3, 4], [4, 0, 0, 0], [7, 0, 0, 0]]).astype(np.float)
    A_transform_excl_2_ax0 = np.array([[1, 2, 3, 4], [4, 0, 0, 0], [7, 8, 9, 10]]).astype(np.float)

    constr = ConstraintCutAbove(copy=True, value=4, exclude=0, exclude_axis=-1)
    out = constr.transform(A)
    assert_allclose(out, A_transform_excl_0_ax1)

    constr = ConstraintCutAbove(copy=True, value=4, exclude=2, exclude_axis=0)
    out = constr.transform(A)
    assert_allclose(out, A_transform_excl_2_ax0)

    # No Copy
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    constr = ConstraintCutAbove(copy=False, value=4, exclude=0, exclude_axis=-1)
    _ = constr.transform(A)
    assert_allclose(A, A_transform_excl_0_ax1)

    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    constr = ConstraintCutAbove(copy=False, value=4, exclude=2, exclude_axis=0)
    _ = constr.transform(A)
    assert_allclose(A, A_transform_excl_2_ax0)
def test_compress_above():
    """ Test compressing above (and not equal to) a value """
    A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(np.float)
    A_transform = np.array([[1, 2, 3, 4], [4, 4, 4, 4], [4, 4, 4, 4]]).astype(np.float)

    constr = ConstraintCompressAbove(copy=True, value=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform)

    # No Copy
    constr = ConstraintCompressAbove(copy=False, value=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform)

def test_replace_zeros():
    """ Test replace zeros using a single feature """
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=0, feature = 0
    A_transform_ax0 = np.array([[0.0, 0.0, 0.0, 1.0],
                                [0.25, 0.25, 0.0, 0.0],
                                [0.3, 0.9, 0.6, 0.0],
                                [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=1, feature = 0
    A_transform_ax1 = np.array([[1.0, 0.0, 0.0, 0.0],
                                [0.25, 0.25, 0.0, 0.0],
                                [0.3, 0.9, 0.6, 0.0],
                                [1.0, 0.0, 0.0, 0.0]], dtype=np.float)

    # Axis 0, copy=True
    constr = ConstraintReplaceZeros(copy=True, axis=0, feature=0)
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax0)

    # Axis 1, copy=True
    constr = ConstraintReplaceZeros(copy=True, axis=1, feature=0)
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax1)

    # Axis 0, copy=FALSE
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=0, feature=0)
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax0)

    # Axis 1, copy=FALSE
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=1, feature=0)
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax1)

def test_replace_zeros_multifeature():
    """ Replace zeros using multiple features """

    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=0, feature = [0,1]
    A_transform_ax0 = np.array([[0.0, 0.0, 0.0, 0.5],
                                [0.25, 0.25, 0.0, 0.5],
                                [0.3, 0.9, 0.6, 0.0],
                                [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=1, feature = [0,1]
    A_transform_ax1 = np.array([[0.5, 0.5, 0.0, 0.0],
                                [0.25, 0.25, 0.0, 0.0],
                                [0.3, 0.9, 0.6, 0.0],
                                [0.5, 0.5, 0.0, 0.0]], dtype=np.float)

    # Axis 0, copy=True, feature=[0,1]
    constr = ConstraintReplaceZeros(copy=True, axis=0, feature=[0,1])
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax0)

    # Axis 1, copy=True, feature=[0,1]
    constr = ConstraintReplaceZeros(copy=True, axis=1, feature=[0,1])
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax1)

    # Axis 1, copy=True, feature=np.array([0,1])
    constr = ConstraintReplaceZeros(copy=True, axis=1, feature=np.array([0,1]))
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax1)

    # Axis 1, copy=True, feature=None
    constr = ConstraintReplaceZeros(copy=True, axis=1, feature=None)
    out = constr.transform(A)
    assert_allclose(out, A)

    # Axis 0, copy=FALSE, feature=[0,1]
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=0, feature=[0,1])
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax0)

    # Axis 1, copy=FALSE, feature=[0,1]
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=1, feature=[0,1])
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax1)

def test_replace_zeros_non1fval():
    """ Test replace zeros using a single feature with fval != 1 """
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=0, feature = 0
    A_transform_ax0 = np.array([[0.0, 0.0, 0.0, 4.0],
                                [0.25, 0.25, 0.0, 0.0],
                                [0.3, 0.9, 0.6, 0.0],
                                [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=1, feature = 0
    A_transform_ax1 = np.array([[4.0, 0.0, 0.0, 0.0],
                                [0.25, 0.25, 0.0, 0.0],
                                [0.3, 0.9, 0.6, 0.0],
                                [4.0, 0.0, 0.0, 0.0]], dtype=np.float)

    # Axis 0, copy=True
    constr = ConstraintReplaceZeros(copy=True, axis=0, feature=0, fval=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax0)

    # Axis 1, copy=True
    constr = ConstraintReplaceZeros(copy=True, axis=1, feature=0, fval=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax1)

    # Axis 0, copy=FALSE
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=0, feature=0, fval=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax0)

    # Axis 1, copy=FALSE
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=1, feature=0, fval=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax1)

def test_replace_zeros_non1fval_multifeature():
    """ Test replace zeros using a 2 features with fval != 1 """
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=0, feature = 0
    A_transform_ax0 = np.array([[0.0, 0.0, 0.0, 2.0],
                                [0.25, 0.25, 0.0, 2.0],
                                [0.3, 0.9, 0.6, 0.0],
                                [0.0, 0.0, 0.0, 0.0]], dtype=np.float)
    # Axis=1, feature = 0
    A_transform_ax1 = np.array([[2.0, 2.0, 0.0, 0.0],
                                [0.25, 0.25, 0.0, 0.0],
                                [0.3, 0.9, 0.6, 0.0],
                                [2.0, 2.0, 0.0, 0.0]], dtype=np.float)

    # Axis 0, copy=True
    constr = ConstraintReplaceZeros(copy=True, axis=0, feature=[0,1], fval=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax0)

    # Axis 1, copy=True
    constr = ConstraintReplaceZeros(copy=True, axis=1, feature=[0,1], fval=4)
    out = constr.transform(A)
    assert_allclose(out, A_transform_ax1)

    # Axis 0, copy=FALSE
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=0, feature=[0,1], fval=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax0)

    # Axis 1, copy=FALSE
    A = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.25, 0.25, 0.0, 0.0],
                  [0.3, 0.9, 0.6, 0.0],
                  [0.0, 0.0, 0.0, 0.0]], dtype=np.float)

    constr = ConstraintReplaceZeros(copy=False, axis=1, feature=[0,1], fval=4)
    out = constr.transform(A)
    assert_allclose(A, A_transform_ax1)

def test_planarize_no_noise():
    """ Test ConstraintPlanarize with no noise """
    C_img = np.zeros((10, 20, 2))  # Y, X, Target
    x = np.arange(C_img.shape[1])
    y = np.arange(C_img.shape[0])
    n_targets = C_img.shape[-1]

    X, Y = np.meshgrid(x,y)

    C_0_ideal = 0.1*X + 0.3*Y + 3
    C_img[:,:,0] = C_0_ideal
    C_img[:,:,1] = 0.2*X + 0.4*Y + 3

    C_ravel = C_img.reshape((-1, n_targets))

    # COPY
    constr = ConstraintPlanarize(0, (10, 20), scaler=None, copy=True)
    out = constr.transform(C_ravel)
    out_img = out.reshape(C_img.shape)

    assert_allclose(out_img[..., 1], C_img[..., 1])
    # Data is a plane. Should return the same plane to within numerical error
    assert_allclose(out_img[...,0], C_0_ideal)

    # OVERWRITE
    constr = ConstraintPlanarize(0, (10, 20), scaler=None, copy=False)
    _ = constr.transform(C_ravel)
    assert_allclose(C_ravel.reshape(C_img.shape)[..., 0], C_img[..., 0])
    assert_allclose(C_ravel.reshape(C_img.shape)[..., 1], C_img[..., 1])


def test_planarize_noisy():
    """ Test ConstraintPlanarize """
    C_img = np.zeros((10, 20, 2))  # Y, X, Target
    x = np.arange(C_img.shape[1])
    y = np.arange(C_img.shape[0])
    n_targets = C_img.shape[-1]

    X, Y = np.meshgrid(x,y)

    C_img[:,:,0] = 0.1*X + 0.3*Y + 3 + 0.1*np.random.randn(10,20)
    C_img[:,:,1] = 0.1*X + 0.3*Y + 3

    C_ravel = C_img.reshape((-1, n_targets))

    constr = ConstraintPlanarize(0, (10, 20), scaler=None, copy=True)
    out = constr.transform(C_ravel)
    out_img = out.reshape(C_img.shape)

    assert_allclose(out_img[..., 1], C_img[..., 1])

    assert np.sum((out_img[..., 0] - C_img[..., 1])**2) < np.sum((C_img[...,0] - C_img[..., 1])**2)

def test_planarize_noisy_list_target():
    """ Test ConstraintPlanarize """
    C_img = np.zeros((10, 20, 2))  # Y, X, Target
    x = np.arange(C_img.shape[1])
    y = np.arange(C_img.shape[0])
    n_targets = C_img.shape[-1]

    X, Y = np.meshgrid(x,y)

    C_img[:,:,0] = 0.1*X + 0.3*Y + 3 + 0.1*np.random.randn(10,20)
    C_img[:,:,1] = 0.1*X + 0.3*Y + 3 + 0.1*np.random.randn(10,20)
    C_ideal_01 = 0.1*X + 0.3*Y + 3

    C_ravel = C_img.reshape((-1, n_targets))

    constr = ConstraintPlanarize([0, 1], (10, 20), scaler=None, copy=True)
    out = constr.transform(C_ravel)
    out_img = out.reshape(C_img.shape)

    # Test that RSS is better after planarize than before
    assert np.sum((C_img[...,0] - C_ideal_01)**2) > np.sum((out_img[...,0] - C_ideal_01)**2)
    assert np.sum((C_img[...,1] - C_ideal_01)**2) > np.sum((out_img[...,1] - C_ideal_01)**2)

    # Two output targets are not identical
    assert np.sum((out_img[..., 0] - out_img[..., 1])**2) > 0

def test_planarize_err_type_input():
    """ Inputting a target that is not a list, tuple, or ndarray results in Type Error """
    with pytest.raises(TypeError):
        constr = ConstraintPlanarize({'a' : 0}, (10, 20), scaler=None, copy=True)

def test_planarize_set_scaler():
    """ Test setting or not setting scaler """

    C_img = np.zeros((10, 20, 2))  # Y, X, Target
    x = np.arange(C_img.shape[1])
    y = np.arange(C_img.shape[0])
    n_targets = C_img.shape[-1]

    X, Y = np.meshgrid(x,y)

    C_img[:,:,0] = 0.1*X + 0.3*Y - 2.5
    C_img[:,:,1] = 0.1*X + 0.3*Y - 2.5

    C_ravel = C_img.reshape((-1, n_targets))

    constr = ConstraintPlanarize(0, (10, 20), scaler=None, copy=True)
    out = constr.transform(C_ravel)

    assert constr.scaler is not None
    assert constr.scaler > 100

    constr = ConstraintPlanarize(0, (10, 20), scaler=1.0, copy=True)
    out = constr.transform(C_ravel)

    assert constr.scaler is not None
    assert constr.scaler == 1.0




def test_planarize_use_above_and_below_on_plane():
    """ Test ConstraintPlanarize """
    C_img = np.zeros((10, 20, 2))  # Y, X, Target
    x = np.arange(C_img.shape[1])
    y = np.arange(C_img.shape[0])
    n_targets = C_img.shape[-1]

    X, Y = np.meshgrid(x,y)

    C_img[:,:,0] = 0.1*X + 0.3*Y - 2.5
    C_img[:,:,1] = 0.1*X + 0.3*Y - 2.5

    C_ravel = C_img.reshape((-1, n_targets))

    # COPY -- DO NOT Apply limits to plane
    constr = ConstraintPlanarize(0, (10, 20), scaler=None, copy=True, use_vals_above=0,
                                 use_vals_below=1, lims_to_plane=False)
    out = constr.transform(C_ravel)

    assert C_ravel.min() < 0
    assert C_ravel.max() > 1
    assert out[:,0].min() < 0
    assert out[:,0].max() > 1
    assert out[:,1].min() < 0
    assert out[:,1].max() > 1

    # COPY -- Apply limits to plane
    constr = ConstraintPlanarize(0, (10, 20), scaler=None, copy=True, use_vals_above=0,
                                 use_vals_below=1, lims_to_plane=True)
    out = constr.transform(C_ravel)

    assert C_ravel.min() < 0
    assert C_ravel.max() > 1
    assert out[:,0].min() >= 0
    assert out[:,0].max() <= 1
    assert out[:,1].min() < 0
    assert out[:,1].max() > 1

    # OVERWRITE
    assert C_ravel[:,0].min() < 0
    assert C_ravel[:,0].max() > 1
    assert C_ravel[:,1].min() < 0
    assert C_ravel[:,1].max() > 1


    constr = ConstraintPlanarize(0, (10, 20), scaler=None, copy=False, use_vals_above=0,
                                 use_vals_below=1, lims_to_plane=True)
    out = constr.transform(C_ravel)

    assert C_ravel[:,0].min() >= 0
    assert C_ravel[:,0].max() <= 1
    assert C_ravel[:,1].min() < 0
    assert C_ravel[:,1].max() > 1



""" Test data conditioning """
import numpy as np

from numpy.testing import (assert_equal, assert_array_equal,
                           assert_allclose)

from pymcr.condition import (standardize)

def test_scale():
    """ Test data standardization """
    A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])

    A_sub_mean_ax0 = np.array([[-1.5, -1.5, -1.5], [1.5, 1.5, 1.5]])
    A_sub_mean_ax1 = np.array([[-1.0, 0.0, 1.0], [-1.0, 0.0, 1.0]])

    A_just_std_ax0 = np.array([[0.66666667, 1.33333333, 2.0], [2.66666667, 3.33333333, 4.0]])
    A_just_std_ax1 = np.array([[1.22474487, 2.44948974, 3.67423461],
                               [4.89897949, 6.12372436, 7.34846923]])

    A_standardized_ax0 = np.array([[-1.0, -1.0, -1.0], [1.0, 1.0, 1.0]])
    A_standardized_ax1 = np.array([[-1.22474487, 0.0, 1.22474487], [-1.22474487, 0.0, 1.22474487]])

    # MEAN-CENTER, COPY, AXIS=0
    out = standardize(A, mean_ctr=True, with_std=False, axis=0, copy=True)
    assert_allclose(out, A_sub_mean_ax0)

    # MEAN-CENTER, COPY, AXIS=1
    out = standardize(A, mean_ctr=True, with_std=False, axis=1, copy=True)
    assert_allclose(out, A_sub_mean_ax1)

    # STD-NORM, COPY, AXIS=0
    out = standardize(A, mean_ctr=False, with_std=True, axis=0, copy=True)
    assert_allclose(out, A_just_std_ax0)

    # STD-NORM, COPY, AXIS=1
    out = standardize(A, mean_ctr=False, with_std=True, axis=1, copy=True)
    assert_allclose(out, A_just_std_ax1)

    # FULL STANDARDIZE, COPY, AXIS=0
    out = standardize(A, mean_ctr=True, with_std=True, axis=0, copy=True)
    assert_allclose(out, A_standardized_ax0)

    # FULL STANDARDIZE, COPY, AXIS=1
    out = standardize(A, mean_ctr=True, with_std=True, axis=1, copy=True)
    assert_allclose(out, A_standardized_ax1)

    # OVERWRITE VARIANTS
    # MEAN-CENTER, OVERWRITE, AXIS=0
    A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    _ = standardize(A, mean_ctr=True, with_std=False, axis=0, copy=False)
    assert_allclose(A, A_sub_mean_ax0)

    # MEAN-CENTER, OVERWRITE, AXIS=1
    A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    _ = standardize(A, mean_ctr=True, with_std=False, axis=1, copy=False)
    assert_allclose(A, A_sub_mean_ax1)

    # STD-NORM, OVERWRITE, AXIS=0
    A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    _ = standardize(A, mean_ctr=False, with_std=True, axis=0, copy=False)
    assert_allclose(A, A_just_std_ax0)

    # STD-NORM, OVERWRITE, AXIS=1
    A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    _ = standardize(A, mean_ctr=False, with_std=True, axis=1, copy=False)
    assert_allclose(A, A_just_std_ax1)

    # FULL STANDARDIZE, OVERWRITE, AXIS=0
    A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    _ = standardize(A, mean_ctr=True, with_std=True, axis=0, copy=False)
    assert_allclose(A, A_standardized_ax0)

    # FULL STANDARDIZE, OVERWRITE, AXIS=1
    A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    _ = standardize(A, mean_ctr=True, with_std=True, axis=1, copy=False)
    assert_allclose(A, A_standardized_ax1)



""" pyMCR: Pythonic Multivariate Curve Resolution - Alternating Least Squares """
from . import mcr
from . import constraints
from . import metrics
from . import regressors
from . import condition

import sys
import logging as _logging

from ._version import __version__

_logger = _logging.getLogger('pymcr')
_logger.setLevel(_logging.DEBUG)

# StdOut is a "stream"; thus, StreamHandler
stdout_handler = _logging.StreamHandler(stream=sys.stdout)

# Set the message format. Simple and removing log level or date info
stdout_format = _logging.Formatter('%(message)s')  # Just a basic message akin to print statements
stdout_handler.setFormatter(stdout_format)

_logger.addHandler(stdout_handler)




"""
Testing external regressors if library if installed

"""

import numpy as np

from numpy.testing import (assert_equal, assert_array_equal, 
                           assert_allclose)

try:
    from sklearn.linear_model import LinearRegression, Ridge
except:
    flag_sklearn = False
else:
    flag_sklearn = True

import pytest   

@pytest.mark.skipif(flag_sklearn == False, reason='sklearn not installed, skipping.')
class TestSklearn:
    """ Test some scikit-learn methods. Skips if sklearn is not installed """
    
    def test_sklearn_linear_regression(self):
        """ Test sklearn linear regression """

        A = np.array([[1,1,0],[1,0,1],[0,0,1]])
        x = np.array([1,2,3])
        X = x[:,None]
        B = np.dot(A,X)
        b = np.dot(A,x)

        # HAVE to NOT fit intercept
        regr = LinearRegression(fit_intercept=False)

        assert hasattr(regr, 'fit')

        regr.fit(A,B)
        assert_allclose(X.T, regr.coef_)

        regr.fit(A,b)
        assert_allclose(x.T, regr.coef_)



"""
Setup for pyMCR
"""
import os
import io
from setuptools import setup, find_packages

with io.open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

with open('./requirements.txt','r') as f:
        INSTALL_REQUIRES = f.read().splitlines()

# get __version__ from _version.py
ver_file = os.path.join('pymcr', '_version.py')
with open(ver_file) as f:
    exec(f.read())

VERSION = __version__  # noqa: F821

setup(name='pyMCR',
      version = VERSION,
      description = 'Multivariate Curve Resolution in Python',
      long_description = long_description,
      url = 'https://github.com/usnistgov/pyMCR',
      author = 'Charles H. Camp Jr.',
      author_email = 'charles.camp@nist.gov',
      license = 'Public Domain',
      packages = find_packages(),
      zip_safe = False,
      include_package_data = True,
      install_requires=INSTALL_REQUIRES,
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 3 :: Only',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Topic :: Scientific/Engineering :: Information Analysis',
                   'Topic :: Scientific/Engineering :: Bio-Informatics',
                   'Topic :: Scientific/Engineering :: Chemistry',
                   'Topic :: Scientific/Engineering :: Information Analysis',
                   'Topic :: Scientific/Engineering :: Mathematics',
                   'Topic :: Scientific/Engineering :: Physics'])



import numpy as np


from numpy.testing import assert_allclose, assert_equal, assert_array_less

import pytest

import pymcr
from pymcr.mcr import McrAR
from pymcr.metrics import mse
from pymcr.constraints import ConstraintNonneg, ConstraintNorm

@pytest.fixture(scope="function")
def dataset():
    """ Setups dataset """

    M = 21
    N = 21
    P = 101
    n_components = 2

    C_img = np.zeros((M,N,n_components))
    C_img[...,0] = np.dot(np.ones((M,1)),np.linspace(0,1,N)[None,:])
    C_img[...,1] = 1 - C_img[...,0]

    St_known = np.zeros((n_components, P))
    St_known[0,40:60] = 1
    St_known[1,60:80] = 2

    C_known = C_img.reshape((-1, n_components))

    D_known = np.dot(C_known, St_known)

    yield C_known, D_known, St_known

def test_mcr_ideal_default(dataset):
    """ Provides C/St_known so optimal should be 1 iteration """

    C_known, D_known, St_known = dataset

    mcrar = McrAR()
    mcrar.fit(D_known, ST=St_known)
    assert_equal(1, mcrar.n_iter_opt)
    assert ((mcrar.D_ - D_known)**2).mean() < 1e-10
    assert ((mcrar.D_opt_ - D_known)**2).mean() < 1e-10

    mcrar.fit(D_known, C=C_known)
    assert_equal(2, mcrar.n_iter_opt)
    assert ((mcrar.D_ - D_known)**2).mean() < 1e-10
    assert ((mcrar.D_opt_ - D_known)**2).mean() < 1e-10

def test_mcr_ideal_str_regressors(dataset):
    """ Test MCR with string-provded regressors"""

    C_known, D_known, St_known = dataset

    mcrar = McrAR(c_regr='OLS', st_regr='OLS')
    mcrar.fit(D_known, ST=St_known, verbose=True)
    assert_equal(1, mcrar.n_iter_opt)
    assert isinstance(mcrar.c_regressor, pymcr.regressors.OLS)
    assert isinstance(mcrar.st_regressor, pymcr.regressors.OLS)

    mcrar = McrAR(c_regr='NNLS', st_regr='NNLS')
    mcrar.fit(D_known, ST=St_known)
    assert_equal(1, mcrar.n_iter_opt)
    assert isinstance(mcrar.c_regressor, pymcr.regressors.NNLS)
    assert isinstance(mcrar.st_regressor, pymcr.regressors.NNLS)
    assert ((mcrar.D_ - D_known)**2).mean() < 1e-10
    assert ((mcrar.D_opt_ - D_known)**2).mean() < 1e-10

    # Provided C_known this time
    mcrar = McrAR(c_regr='OLS', st_regr='OLS')
    mcrar.fit(D_known, C=C_known)

    # Turns out some systems get it in 1 iteration, some in 2
    # assert_equal(1, mcrar.n_iter_opt)
    assert_equal(True, mcrar.n_iter_opt<=2)

    assert ((mcrar.D_ - D_known)**2).mean() < 1e-10
    assert ((mcrar.D_opt_ - D_known)**2).mean() < 1e-10

def test_mcr_max_iterations(dataset):
    """ Test MCR exits at max_iter"""

    C_known, D_known, St_known = dataset

    # Seeding with a constant of 0.1 for C, actually leads to a bad local
    # minimum; thus, the err_change gets really small with a relatively bad
    # error. The tol_err_change is set to None, so it makes it to max_iter.
    mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_increase=None, tol_n_increase=None,
                    tol_err_change=None, tol_n_above_min=None)
    mcrar.fit(D_known, C=C_known*0 + 0.1)
    assert mcrar.exit_max_iter_reached

def test_mcr_tol_increase(dataset):
    """ Test MCR exits due error increasing above a tolerance fraction"""

    C_known, D_known, St_known = dataset

    # Seeding with a constant of 0.1 for C, actually leads to a bad local
    # minimum; thus, the err_change gets really small with a relatively bad
    # error.
    mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_increase=0, tol_n_increase=None,
                    tol_err_change=None, tol_n_above_min=None)
    mcrar.fit(D_known, C=C_known*0 + 0.1)
    assert mcrar.exit_tol_increase

def test_mcr_tol_n_increase(dataset):
    """
    Test MCR exits due iterating n times with an increase in error

    Note: On some CI systems, the minimum err bottoms out; thus, tol_n_above_min
    needed to be set to 0 to trigger a break.
    """

    C_known, D_known, St_known = dataset

    mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_increase=None, tol_n_increase=0,
                    tol_err_change=None, tol_n_above_min=None)
    mcrar.fit(D_known, C=C_known*0 + 0.01)
    assert mcrar.exit_tol_n_increase

def test_mcr_tol_err_change(dataset):
    """ Test MCR exits due error increasing by a value """

    C_known, D_known, St_known = dataset

    mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_increase=None, tol_n_increase=None,
                    tol_err_change=1e-20, tol_n_above_min=None)
    mcrar.fit(D_known, C=C_known)
    assert mcrar.exit_tol_err_change

def test_mcr_tol_n_above_min(dataset):
    """
    Test MCR exits due to half-terating n times with error above the minimum error.

    Note: On some CI systems, the minimum err bottoms out; thus, tol_n_above_min
    needed to be set to 0 to trigger a break.
    """

    C_known, D_known, St_known = dataset

    mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_increase=None, tol_n_increase=None,
                    tol_err_change=None, tol_n_above_min=0)
    mcrar.fit(D_known, C=C_known*0 + 0.1)
    assert mcrar.exit_tol_n_above_min


def test_mcr_st_semilearned():
    """ Test when St items are fixed, i.e., enforced to be the same as the input, always """

    M = 21
    N = 21
    P = 101
    n_components = 3

    C_img = np.zeros((M,N,n_components))
    C_img[...,0] = np.dot(np.ones((M,1)),np.linspace(0,1,N)[None,:])
    C_img[...,1] = np.dot(np.linspace(0,1,M)[:, None], np.ones((1,N)))
    C_img[...,2] = 1 - C_img[...,0] - C_img[...,1]
    C_img = C_img / C_img.sum(axis=-1)[:,:,None]

    St_known = np.zeros((n_components, P))
    St_known[0,30:50] = 1
    St_known[1,50:70] = 2
    St_known[2,70:90] = 3
    St_known += 1

    C_known = C_img.reshape((-1, n_components))

    D_known = np.dot(C_known, St_known)

    ST_guess = 1 * St_known
    ST_guess[2, :] = np.random.randn(P)

    mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_err_change=1e-10)

    mcrar.fit(D_known, ST=ST_guess, st_fix=[0,1])
    assert_equal(mcrar.ST_[0,:], St_known[0,:])
    assert_equal(mcrar.ST_[1,:], St_known[1,:])

def test_mcr_c_semilearned():
    """ Test when C items are fixed, i.e., enforced to be the same as the input, always """

    M = 21
    N = 21
    P = 101
    n_components = 3

    C_img = np.zeros((M,N,n_components))
    C_img[...,0] = np.dot(np.ones((M,1)),np.linspace(0,1,N)[None,:])
    C_img[...,1] = np.dot(np.linspace(0,1,M)[:, None], np.ones((1,N)))
    C_img[...,2] = 1 - C_img[...,0] - C_img[...,1]
    C_img = C_img / C_img.sum(axis=-1)[:,:,None]

    St_known = np.zeros((n_components, P))
    St_known[0,30:50] = 1
    St_known[1,50:70] = 2
    St_known[2,70:90] = 3
    St_known += 1

    C_known = C_img.reshape((-1, n_components))

    D_known = np.dot(C_known, St_known)

    C_guess = 1 * C_known
    C_guess[:, 2] = np.abs(np.random.randn(int(M*N))+0.1)

    mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_err_change=1e-10)

    mcrar.fit(D_known, C=C_guess, c_fix=[0,1])
    assert_equal(mcrar.C_[:, 0], C_known[:, 0])
    assert_equal(mcrar.C_[:, 1], C_known[:, 1])

def test_mcr_semilearned_both_c_st():
    """
    Test the special case when C & ST are provided, requiring C-fix ST-fix to
    be provided
    """

    M = 21
    N = 21
    P = 101
    n_components = 3

    C_img = np.zeros((M,N,n_components))
    C_img[...,0] = np.dot(np.ones((M,1)),np.linspace(0.1,1,N)[None,:])
    C_img[...,1] = np.dot(np.linspace(0.1,1,M)[:, None], np.ones((1,N)))
    C_img[...,2] = 1 - C_img[...,0] - C_img[...,1]
    C_img = C_img / C_img.sum(axis=-1)[:,:,None]

    St_known = np.zeros((n_components, P))
    St_known[0,30:50] = 1
    St_known[1,50:70] = 2
    St_known[2,70:90] = 3
    St_known += 1

    C_known = C_img.reshape((-1, n_components))

    D_known = np.dot(C_known, St_known)

    C_guess = 1 * C_known
    C_guess[:, 2] = np.abs(np.random.randn(int(M*N)))

    mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                    st_constraints=[ConstraintNonneg()],
                    c_constraints=[ConstraintNonneg(), ConstraintNorm()],
                    tol_err_change=1e-10)

    mcrar.fit(D_known, C=C_guess, ST=St_known, c_fix=[0,1], st_fix=[0], c_first=True)
    assert_equal(mcrar.C_[:, 0], C_known[:, 0])
    assert_equal(mcrar.C_[:, 1], C_known[:, 1])
    assert_equal(mcrar.ST_[0, :], St_known[0, :])

    # ST-solve first
    mcrar.fit(D_known, C=C_guess, ST=St_known, c_fix=[0,1], st_fix=[0], c_first=False)
    assert_equal(mcrar.C_[:, 0], C_known[:, 0])
    assert_equal(mcrar.C_[:, 1], C_known[:, 1])
    assert_equal(mcrar.ST_[0, :], St_known[0, :])

def test_mcr_errors():

    # Providing both C and S^T estimates without C_fix and St_fix
    with pytest.raises(TypeError):
        mcrar = McrAR()
        mcrar.fit(np.random.randn(10,5), C=np.random.randn(10,3),
                   ST=np.random.randn(3,5))

    # Providing both C and S^T estimates without both C_fix and St_fix
    with pytest.raises(TypeError):
        mcrar = McrAR()
        # Only c_fix
        mcrar.fit(np.random.randn(10,5), C=np.random.randn(10,3),
                   ST=np.random.randn(3,5), c_fix=[0])

    with pytest.raises(TypeError):
        mcrar = McrAR()
        # Only st_fix
        mcrar.fit(np.random.randn(10,5), C=np.random.randn(10,3),
                   ST=np.random.randn(3,5), st_fix=[0])

    # Providing no estimates
    with pytest.raises(TypeError):
        mcrar = McrAR()
        mcrar.fit(np.random.randn(10,5))

    # Unknown regression method
    with pytest.raises(ValueError):
        mcrar = McrAR(c_regr='NOTREAL')

    # regression object with no fit method
    with pytest.raises(ValueError):
        mcrar = McrAR(c_regr=print)

def test_props_features_samples_targets(dataset):
    """ Test mcrar properties for features, targets, samples """
    C_known, D_known, St_known = dataset

    mcrar = McrAR()
    mcrar.fit(D_known, ST=St_known)

    assert mcrar.n_targets == C_known.shape[-1]  # n_components
    assert mcrar.n_samples == D_known.shape[0]
    assert mcrar.n_features == D_known.shape[-1]